package Gear::Rules;
use strict;
use warnings;
#use File::Path qw(make_path);
use File::Temp qw/ tempfile tempdir /;
use File::Basename;
#use IPC::Run3;
use Text::ParseWords;
use Carp;

our $VERSION = '0.201';
our $verbose;
our @LONGOPT = (
    );

# static method
sub is_gear_repository_dir {
    my ($dir)=@_;
    $dir//='';
    $dir .='/' if length($dir) and substr($dir,0,1) ne '/';
    return (-d $dir.'.git') && (-f $dir.'.gear-rules' or -f $dir.'.gear/rules');
}

sub new {
    my $proto=shift;
    my $class=ref($proto) || $proto;
    my $self = {
	# -use_absolute_path -- helps with chdir
	-nvr_cache => {},
	@_,
    };
    bless($self,$class);
    unless (defined $self->{-topdir}) {
	my $topdir=Gear::Git::Helper::get_git_toplevel();
	$self->{-topdir}=$self->{-use_absolute_path} ? `pwd`.'/'.$topdir : $topdir;
    }
    $self->__load($self->{'-gear_rules_file'});
    return $self;
}

sub get_top_dir {
    my ($self)=@_;
    return $self->{-topdir};
}

sub get_spec {
    my ($self)=@_;
    my $spec=$self->{'__def_spec__'};
    return $spec if defined $spec;
    $spec=$self->{'spec:'};
    unless (defined $spec) {
	my @spec=glob($self->get_top_dir . "*.spec");
	$spec=$spec[0] if @spec==1;
	die "Gear::Rules: too much spec files! ".join(' ',@spec) if @spec>1;
	die "Gear::Rules: spec file not found!" if @spec<1;
    }
    $self->{'__def_spec__'}=$spec;
    return $spec;
}

sub get_rules {
    my ($self)=@_;
    return @{$self->{RULES}};
}

sub get_specsubst {
    my ($self)=@_;
    return $self->{'specsubst:'};
}

sub get_nvr {
    my ($self,$specfile)=@_;
    my $def_specfile=$self->get_spec();
    $specfile||=$def_specfile;
    my $nvrptr=$self->{-nvr_cache}->{$specfile};
    return @$nvrptr if $nvrptr;
    my @nvr;
    if ($def_specfile eq $specfile) {
	@nvr=&__get_gear_nvr();
    } else {
	@nvr=&__get_gear_nvr($specfile);
    }
    $self->{-nvr_cache}->{$specfile}=\@nvr;
    return @nvr;
}

sub get_gear_tag_commit {
    my ($self, $tag)=@_;
    return $self->{GEAR_TAGS}->{$tag};
}

sub get_tagsdir {
    my ($self)=@_;
    my $topdir=$self->get_top_dir;
    my $tagsdir=$self->{'tags:'};
    $tagsdir||=$topdir.'.gear-tags' if -d $topdir.'.gear-tags';
    $tagsdir||=$topdir.'.gear/tags' if -d $topdir.'.gear/tags';
    return $tagsdir;
}

sub has_external_commits {
    my ($self)=@_;
    my $has_external_commits=$self->{__cached_has_external_commits};
    return $has_external_commits if defined ($has_external_commits);
    $has_external_commits=$self->__has_external_commits;
    $self->{__cached_has_external_commits}=$has_external_commits;
    return $has_external_commits;
}

sub external_commits {
    my ($self)=@_;
    my $external_commits=$self->{__cached_external_commits};
    return $external_commits if defined ($external_commits);
    $external_commits=[];
    $self->__has_external_commits($external_commits);
    $self->{__cached_external_commits}=$external_commits;
    return $external_commits;
}

sub __load {
    my ($self, $gearrules)=@_;
    my $topdir=$self->get_top_dir;
    $gearrules ||= $topdir.'.gear-rules' if -e $topdir.'.gear-rules';
    $gearrules ||= $topdir.'.gear/rules' if -e $topdir.'.gear/rules';

    $self->{RULES}=[];
    if ($gearrules) {
	$self->{'gear-rules:'}=$gearrules;
	open (GEARRULES, '<', $gearrules) || die "can't open $gearrules:$!";
	my $lineno=0;
	local $_;
	while (<GEARRULES>) {
	    chomp;
	    my $string=$_;
	    $lineno++;
	    &__process_rules_line($self, $string, $lineno);
	}
	close (GEARRULES);
    } elsif (-d $topdir.'.gear/tags' or -d $topdir.'.gear-tags') {
	# implicit gear rule
	$self->{'gear-rules:'}='(gear rules are missing: default is tar .)';
	&__process_rules_line($self, 'tar: .', 1);
    } else {
	die "Not a gear repository: gear rules file not found.\n" unless defined $gearrules;
    }

    $self->__load_gear_tags();
}

sub __process_rules_line {
    my ($self, $string, $lineno)=@_;
    my $gearrules=$self->{'gear-rules:'};
    my $topdir=$self->get_top_dir;
    local $_;
    $_=$string;
    if (/^\s*(#|$)/) {
	return;
    } elsif (/^\s*([^:]+):\s*(.+)$/) {
	#print "found cmd $1: $2\n";
	my $cmd=$1;
	my $args=$2;
	$args=~s/\s*$//;
	my $optional=0;
	if ($cmd eq 'spec') {
	    $self->{$cmd.':'}=$args;
	    warn "invalid spec: directive: file $args does not exist" unless -e $topdir.$args;
	} elsif ($cmd eq 'tags' or $cmd eq 'exclude') {
	    $self->{$cmd.':'}=$args;
	} elsif ($cmd eq 'specsubst') {
	    my @specsubst_varlist=split(/\s+/,$args);
	    $self->{$cmd.':'}=\@specsubst_varlist;
	} elsif ($cmd =~/^(copy|compress|gzip|bzip2|lzma|xz)(\?)?$/ ) {
	    my $cmd=$1;
	    $optional=1 if $2;
	    my ($inpatternptr,$opts,$zipoptptr)=&__parse_gear_gzip_args($self,$args);
	    my @outpatterns=map {&__in_to_out_pattern($opts->{'type'} ? $opts->{'type'}:$cmd,undef,$_)} @$inpatternptr;
	    @outpatterns=[$opts->{'name'}] if $opts->{'name'};
	    my @inpatterns=map {['.',$_]} @$inpatternptr;
	    push @{$self->{RULES}}, Gear::Rule->new(-rules=>$self,-line=>$lineno,-cmd=>$cmd,-optional=>$optional, -inpatterns=>\@inpatterns, -outpatterns=>\@outpatterns, -rulestring=>$string, -opts=>$opts, -zip_args=>$zipoptptr, -type=>'copy');
	    
	} elsif ($cmd =~/^diff(\.(?:gz|bz2|lzma|xz))?(\?)?$/ ) {
	    my $cmd='diff';
	    $cmd='diff'.$1 if $1;
	    $optional=1 if $2;
	    #
	    my ($tree_path1, $tree_path2, $opts)=&__parse_gear_diff_args($self,$args);
	    my ($tree1,$path1)=&split_tree_path($tree_path1);
	    my ($tree2,$path2)=&split_tree_path($tree_path2);
	    my $inpatternptr=[[$tree1, $path1], [$tree2, $path2]];
	    my $outstem='@new_dir@-@version@-@release@.patch';
	    $outstem='@name@-@version@-@release@.patch' if $path2 eq '.';
	    $outstem=$opts->{'name'} if $opts->{'name'};
	    my $_dir_=basename($path1);
	    $outstem=~s/\@old_dir\@/$_dir_/g;
	    $_dir_=basename($path2);
	    $outstem=~s/\@new_dir\@/$_dir_/g;
	    my $outpattern=&__in_to_out_pattern($cmd,undef,$outstem);
	    $outpattern=$outstem if $opts->{'name'};
	    
	    push @{$self->{RULES}}, Gear::Rule->new(-rules=>$self,-line=>$lineno,-cmd=>$cmd,-optional=>$optional, -inpatterns=>$inpatternptr, -outpatterns=>[$outpattern], -rulestring=>$string, -opts=>$opts, -tree1=>$tree1, -path1=>$path1, -tree2=>$tree2, -path2=>$path2, -type=>'diff');

	} elsif ($cmd =~/^(tar|tar.bz2|tar.gz|tar.lzma|tar.xz|zip|jar)(\?)?$/ ) {
	    my $cmd=$1;
	    $optional=1 if $2;

	    my ($tree_path,$opts)=&__parse_gear_tar_args($self,$args);
	    my ($tree,$dir)=&split_tree_path($tree_path);
	    my $inpatternptr=[[$tree, $dir]];
	    my $outstem='@dir@-@version@';
	    $outstem='@name@-@version@' if $dir eq '.';
	    $outstem=$opts->{'name'} if $opts->{'name'};
	    # should calculate, not store
	    #$opts->{'base'}=$opts->{'name'} unless $opts->{'base'};
	    if ($dir ne '.') {
		my $_dir_=basename($dir);
		$outstem=~s/\@dir\@/$_dir_/g;
	    }
	    my $outpattern=&__in_to_out_pattern($cmd,$opts->{'suffix'},$outstem);
	    push @{$self->{RULES}}, Gear::Rule->new(-rules=>$self,-line=>$lineno,-cmd=>$cmd,-optional=>$optional, -inpatterns=>$inpatternptr, -outpatterns=>[$outpattern], -rulestring=>$string, -opts=>$opts, -tree=>$tree, -path=>$dir, -type=>'tar');
	} else {
	    warn "$gearrules: found unsupported directive $cmd: $args\n";
	}
    } else {
	warn "$gearrules: found unrecognized line $_\n";
    }
}


sub match_srpm_file {
    my ($rules,$filestring,$gearspecfile)=@_;
    return unless $filestring;
    # some rules can have alternative spec file. By default, we use main specfile.
    $gearspecfile||=$rules->get_spec();
    $gearspecfile=&__trim_filename($gearspecfile);
    my $found;
    my $srcfile=basename($filestring);
    my @nvr=$rules->get_nvr($gearspecfile);
    my $topdir=$rules->get_top_dir;

  RULE: 
    # &reverse() because the last rule will overwrite the file
    foreach my $rule (reverse($rules->get_rules)) {
	my $opts=$rule->{-opts};
	# the rule refer to other specfile
	next if $opts and defined $opts->{'spec'} and &__trim_filename($opts->{'spec'}) ne $gearspecfile;
	foreach my $glob (@{$rule->{-outpatterns}}) {
	    my $outpattern_expanded=&subst_gear_nvr($glob,@nvr);
	    my $regexp=&glob2pat($outpattern_expanded);
	    #print STDERR "glob2pat: $glob -> $outpattern_expanded -> $regexp\n";
	    if ($srcfile=~/$regexp/) {
		$found=$rule;
		last RULE;
	    } else {
		# TODO add checks to be correct:
		# 1) chould check type copy(?), gzip(?), ...
		# 2) chould check exclude:
		my $patterndir=dirname($outpattern_expanded);
		if (defined($patterndir) and $patterndir ne '' and -d $topdir.$patterndir) {
		    $regexp=&glob2pat(basename($outpattern_expanded));
		    #print STDERR "glob2pat2: $glob -> $regexp\n";
		    if ($srcfile=~/$regexp/) {
			$found=$rule;
			last RULE;
		    }
		}
	    }
	}
    }
    return $found;
}

sub __load_gear_tags {
    my ($self)=@_;
    my %tags;
    my $tagsdir=$self->get_tagsdir;
    if ($tagsdir and -d $tagsdir and -e "$tagsdir/list") {
	open (TAGLIST, '<', "$tagsdir/list") or die "can't open $tagsdir/list:$!";
	while (<TAGLIST>) {
	    chomp;
	    if (/^(\S+)\s+(\S.*)$/) {
		my $tagname=$2;
		my $object=$1;
		my $type;
		($type, $object)=&__resolve_gear_commit($tagsdir,$tagname,$object);
		$tags{$tagname}=$object;
	    }
	}
	close (TAGLIST);
    }
    $self->{GEAR_TAGS}=\%tags;
}


sub __resolve_gear_commit {
    my ($tagsdir,$tagname,$object)=@_;
    # 2>/dev/null: fatal: git cat-file: could not get object info
    my $type=`git cat-file -t $object 2>/dev/null`;
    chomp $type;
    if ($?) {
	# support for lost tags
	my $newobj;
	if (-e "$tagsdir/$object") {
	    open (TAGENTRY, '<', "$tagsdir/$object") or die "can't open $tagsdir/$object:$!";
	    while (<TAGENTRY>) {
		chomp;
		if (/^object\s+(\S.*)$/) {
		    $newobj=$1;
		} elsif (/^type\s+(\S.*)$/) {
		    $type=$1;
		}
	    }
	    close (TAGENTRY);
	    if ($newobj) {
		($type, $object)=&__resolve_gear_commit($tagsdir,$tagname,$newobj);
		&__check_commit_type($type, "$tagsdir/list error: in tag $tagname: $object points to a non-commit type git object $newobj\n");
	    } else {
		die "Gear::Rules: bad gear list. Tag $tagname and git object $object are not present in this git repository. Try to fetch them somewhere or edit $tagsdir/list.\n";
	    }
	}
    }
    &__check_commit_type($type, "$tagsdir/list error: Tag $tagname points to a non-commit type git object $object\n");
    return ($type, $object);
}

sub __check_commit_type {
    my ($type,$message)=@_;
    die $message if $type eq 'blob' or $type eq 'tree';
}

### use git log format
my $ignore_external_commits_name='.gear/uupdate_ignore_commits';

sub __has_external_commits {
    my ($self,$external_commits_collector_ref)=@_;
    my $ignore_external_commits_file=$self->get_top_dir.$ignore_external_commits_name;
    my %ignore_commits;
    local $_;
    if (-r $ignore_external_commits_file) {
	open (my $ignore_fn, '<', $ignore_external_commits_file) || die "can't open $ignore_external_commits_file: $!";
	while(<$ignore_fn>) {
	    chomp;
	    if (/^commit\s+([0-9a-f]+)$/) {
		$ignore_commits{$1}=1;
	    }
	}
	close ($ignore_fn);
    }
    my $gitlog_fn;
    open ($gitlog_fn, 'git log |') || die "can't exec git log: $!";
    my $has_external_commits;
    my $commit;
    my $can_skip_further_logs;
    while(<$gitlog_fn>) {
	chomp;
	# skip logs
	next if $can_skip_further_logs;
	#print STDERR "Rules: external_commits: $_\n";
	if (/^commit\s+([0-9a-f]+)$/) {
	    $commit=$1;
	} elsif (/^Author:\s+(\S.*)$/) {
	    my $author=$1;
	    if ($author!~/(
		ab\@samba\.org|
		thecrux\@gmail\.com|
		lnkvisitor\@lnkvisitor\.localdomain|
		altlinux|etersoft|hasher).*>?$/x
		and not $ignore_commits{$commit}
		) {
		$has_external_commits=1;
		if ($external_commits_collector_ref) {
		    push @$external_commits_collector_ref, $commit;
		} else {
		    $can_skip_further_logs=1;
		}
	    }
	}
    }
    close($gitlog_fn);
    return 1 if $has_external_commits;
}




my $merge_threshhold=.8;
sub __guess_merge_strategy {
    my ($rules,$commit1,$commit2)=@_;
    confess "Oops: internal error: __guess_merge_strategy: commit1, commit2 is required" unless defined($commit1) and defined($commit2);
    my ($count1,$files1)=&__ls_git_tree($commit1);
    my ($count2,$files2)=&__ls_git_tree($commit2);
    return 'ours' if $count1*$merge_threshhold > $count2;
    # compare files
    my $filecount=0;
    map {$filecount++ if $files2->{$_}} keys (%$files1);
    return if $filecount > $count1*$merge_threshhold;
    # TODO: compare subtrees to find subtree merge
    return 'ours';
}

sub __ls_git_tree {
    my ($tree)=@_;
    my %files;
    my $count;
    open (my $fh, "git ls-tree -r --name-only $tree |") or die "git ls-tree -r --name-only $tree failed: $!";
    while (<$fh>) {
	chomp;
	$files{$_}=1;
	$count++;
    }
    close($fh);
    return $count, \%files;
}

sub __expand_nvr_in_rule_string {
    my ($rules,$rule,$rstring)=@_;
    my $opts=$rule->{-opts};
    my $rulespecfile;
    # the rule can refer to other specfile
    $rulespecfile=$opts->{'spec'} if $opts;
    my @nvr=$rules->get_nvr($rulespecfile);
    return &__subst_nvr($rstring,@nvr);
}

sub __subst_nvr {
    my ($val,$name,$version,$release)=@_;
    $val=~s/\@name\@/$name/g;
    $val=~s/\@version\@/$version/g;
    $val=~s/\@release\@/$release/g;
    return $val;
}

sub __is_gear_layout_not_robot_recognizeable {
    my ($rules,$source0rule)=@_;
    return 'rule for Source0 is not an archive rule' unless not $source0rule or $source0rule->type eq 'tar';
    my $topdir=$rules->get_top_dir;
    my $diff_subtree_rule_detected=0;
    my (@diff_rule, @branch_tar_rule, %diff_tree);
    foreach my $a_rule ($rules->get_rules()) {
	if ($a_rule->type eq 'diff') {
	    push @diff_rule, $a_rule;
	    $diff_tree{$a_rule->{-path1}}=1;
	    $diff_tree{$a_rule->{-path2}}=1;
	} elsif ($a_rule->type eq 'tar' and $a_rule->{-tree} ne '.') {
	    push @branch_tar_rule, $a_rule;
	}
    }

    foreach my $diffrule (@diff_rule) {
	return 'in diff rule: '.$diffrule->rulestring().
	    ' diferrent paths on the same tree - kind of strange. - not supported yet'
	    if $diffrule->{-tree1} eq $diffrule->{-tree2};
	return 'in diff rule: '.$diffrule->rulestring().
	    ' subtree merge to non-master branch - not supported yet'
	    if $diffrule->{-tree2} ne '.' and
	    $diffrule->{-path1} ne $diffrule->{-path2};

	$diff_subtree_rule_detected=1 if 
	    $diffrule->{-tree2} eq '.' and
	    $diffrule->{-path1} ne $diffrule->{-path2};
    }

    if ($source0rule and $source0rule->{-tree} eq '.') {
	return 'path '.$source0rule->{-path}.' is not found' unless -d $topdir.&__expand_nvr_in_rule_string($rules,$source0rule,$source0rule->{-path});
	# looks like upstream repository; can't be updated
	return q{looks like upstream repository: path in tree=`.'. Change your .gear/rules. Instead of `tar: .' use `tar: @version@:.' or `tar: upstream:.' and, if necessary, add `diff: @version@:. .'.} if $source0rule->{-path} eq '.';
	return 'no upstream branches are expected' if @branch_tar_rule>0;
	return;
    }

    # only one upstream branch is expected
    return 'only one upstream branch is expected' if @branch_tar_rule>1;
    return 'diff rules are expected to have the same path in tree unless it is subtree merge to the main branch' if scalar keys(%diff_tree)>1 and not $diff_subtree_rule_detected;
    return;
}



# returns topologically sorted (i.e. in valid order of execution)
# list of diff rules
# http://en.wikipedia.org/wiki/Topological_sorting

# L ← Empty list that will contain the sorted nodes
# S ← Set of all nodes with no incoming edges
# for each node n in S do
#     visit(n) 
# function visit(node n)
#     if n has not been visited yet then
#         mark n as visited
#         for each node m with an edge from n to m do
#             visit(m)
#         add n to L

# There is an edge from x to y if job x depends on job y

# for diffs, it means edge from x to y if old1tree(x) = new2tree(y)
# incoming edge means new2tree is used as old1tree somewhere

sub sorted_diff_rules {
    my ($rules)=@_;
    my @diff_rules_unsorted = grep {$_->type eq 'diff'} $rules->get_rules();
    my %incoming_edges;
    map {$incoming_edges{&__edge1_id($_)}=1} @diff_rules_unsorted;
    my @no_incoming_edges=grep {not $incoming_edges{&__edge2_id($_)}} @diff_rules_unsorted;
    my @diff_rules_sorted;
    my %visited;
    my $node_n;
    while (@no_incoming_edges) {
	my $node_n = shift @no_incoming_edges;
	&__visit($node_n,\%visited,\@diff_rules_unsorted,\@diff_rules_sorted);
    }
    return @diff_rules_sorted;
}

sub __visit {
    my ($node_n,$visited,$diff_rules_unsorted,$diff_rules_sorted)=@_;
    return if $visited->{$node_n};
    #print '__visiting ',$node_n->rulestring(),"\n";
    $visited->{$node_n}=1;
    my $edge1_n=&__edge1_id($node_n);
    foreach my $node_m (@$diff_rules_unsorted) {
	#print "trying ",$node_m->rulestring(), ' with edge ',&__edge2_id($node_m),' against ',$edge1_n, "\n";
	&__visit($node_m,$visited,$diff_rules_unsorted,$diff_rules_sorted) if $edge1_n eq &__edge2_id($node_m);
    }
    #print 'push ',$node_n->rulestring(),"\n";
    push @$diff_rules_sorted, $node_n;
}

sub __edge1_id {
    my $diffrule=shift;
    return $diffrule->{-tree1}."[\n]".$diffrule->{-path1};
}

sub __edge2_id {
    my $diffrule=shift;
    return $diffrule->{-tree2}."[\n]".$diffrule->{-path2};
}

### static helpers

### protected static helpers ####

sub template_is_likely_a_tag {
    my ($tag_template)=@_;
    return 1 if $tag_template=~/\@(?:version|release)\@/;
    return 0;
}

sub subst_gear_nvr {
    my $string=shift;
    foreach my $key ('name','version','release') {
	my $val=shift;
	$string=~s/\@$key\@/$val/g;
    }
    return $string;
}


### private helpers ####

sub __get_gear_nvr {
    my $specfile=shift;
    my @nvr;
    if (defined $specfile) {
	unless (-e $specfile) {
	    Carp::carp "spec file $specfile not found";
	    return;
	}
	my $tmpfilename = $specfile.'.tmp.gear.rules.'.$$;
	open (my $fh, '>', $tmpfilename) || die "can't create temporary file $tmpfilename: $!";
	print $fh "spec: $specfile\n";
	close ($fh);
	@nvr=split(/\s+/,`gear --describe --disable-specsubst -r $tmpfilename`);
	unlink ($tmpfilename);
    } else {
	@nvr=split(/\s+/,`gear --describe --disable-specsubst`);
    }
    return @nvr[0..2];
}

sub glob2pat {
    my $globstr = shift;
    my %patmap = (
	'*' => '.*',
	'?' => '.',
	'[' => '[',
	']' => ']',
	'{' => '(?:',
	'}' => ')',
	',' => '|',
	);
    $globstr =~ s{(.)} { $patmap{$1} || "\Q$1" }ge;
    return '^' . $globstr . '$';
}

# sub tree_path_to_path {
#     my ($path)=@_;
#     $path=~s/^[^:]://;
#     return $path;
# }

sub split_tree_path {
    my ($path)=@_;
    my $tree='.';
    if ($path=~s/^([^:]+)://) {
	$tree=$1;
    }
    return $tree,$path;
}

sub __trim_filename {
    my $filename=shift;
    $filename=~s!^./!!;
    return $filename;
}

sub __parse_gear_gzip_args {
    my $self=shift;
    my @args=split(/\s+/,shift());
    my (@pattern,%compress_opts, @zip_opts);
    my $endopt=0;
    foreach (@args) {
	if (/^--/) {
	    $endopt=1;
	} elsif (/^-/ and not $endopt) {
	    push @zip_opts, $_;
	} elsif (/^(name|type)=(.*)/) {
	    $compress_opts{$1}=$2;
	} else {
	    push @pattern, $_;
	}
    }
    return \@pattern, \%compress_opts, \@zip_opts;
}

sub __in_to_out_pattern {
    my ($cmd,$suffix,$inpattern)=@_;
    my $pattern=$inpattern;
    if (defined $suffix) {
	$pattern=$inpattern.$suffix;
    } elsif ($cmd =~/^zip$/) {
	$pattern=$inpattern.".zip";
    } elsif ($cmd =~/^tar$/) {
	$pattern=$inpattern.".tar";
    } elsif ($cmd =~/tar\.gz$/) {
	$pattern=$inpattern.".tar.gz";
    } elsif ($cmd =~/tar\.bz2$/) {
	$pattern=$inpattern.".tar.bz2";
    } elsif ($cmd =~/tar\.lzma$/) {
	$pattern=$inpattern.".tar.lzma";
    } elsif ($cmd =~/tar\.xz$/) {
	$pattern=$inpattern.".tar.xz";
    } elsif ($cmd =~/(?:gzip|\.gz)$/) {
	$pattern=$inpattern.".gz";
    } elsif ($cmd =~/(?:bzip2|\.bz2)$/) {
	$pattern=$inpattern.".bz2";
    } elsif ($cmd =~/lzma$/) {
	$pattern=$inpattern.".lzma";
    } elsif ($cmd =~/xz$/) {
	$pattern=$inpattern.".xz";
    }
    return $pattern;
}

sub __parse_gear_tar_args {
    my $self=shift;
    my @args=shellwords(shift());
    my $treepath=shift @args;
    my %opts;
    foreach (@args) {
	if (/^([\w_]+)=(.*)/) {
	    $opts{$1}=$2;
	    if ($1 !~ '^(?:name|base|spec|suffix)$') {
		warn "Gear::Rules: unsupported tar option $_\n";
	    }
	} else {
	    warn "Gear::Rules: unsupported tar argument $_\n";
	}
    }
    my $topdir=$self->get_top_dir;
    warn "Gear::Rules: in tar: spec file $opts{'spec'} does not exist!" if $opts{'spec'} and not -e $topdir.$opts{'spec'};
    return ($treepath,\%opts);
}

sub __parse_gear_diff_args {
    my $self=shift;
    my @args=shellwords(shift());
    my $treepath1=shift @args;
    my $treepath2=shift @args;
    my %opts;
    foreach (@args) {
	if (/^([\w_]+)=(.*)/) {
	    $opts{$1}=$2;
	    if ($1 !~ '^(?:name|spec)$') {
		warn "Gear::Rules: unsupported diff option $_\n";
	    }
	} else {
	    warn "Gear::Rules: unsupported diff argument $_\n";
	}
    }
    my $topdir=$self->get_top_dir;
    warn "Gear::Rules: in diff: spec file $opts{'spec'} does not exist!" if $opts{'spec'} and not -e $topdir.$opts{'spec'};
    return ($treepath1, $treepath2,\%opts);
}

package Gear::Rule;

# read-only object, no setters

sub new {
    my $proto=shift;
    my $class=ref($proto) || $proto;
    my $self = {
	@_,
    };
    bless($self,$class);
    return $self;
}

sub type {
    my $rule=shift;
    return $rule->{-type};
}

sub linenum {
    my $rule=shift;
    return $rule->{-line};
}

sub rules {
    my $rule=shift;
    return $rule->{-rules};
}

sub rulestring {
    my $rule=shift;
    return $rule->{-rulestring};
}

sub specfile {
    my $rule=shift;
    my $type=$rule->type;
    if (($type eq 'tar' or $type eq 'diff') and $rule->{-opts}) {
	my $specfile=$rule->{-opts}->{spec};
	return $specfile if $specfile;
    }
    return $rule->{-rules}->get_spec();
}

sub git_paths {
    my $rule=shift;
    my $patternptr=$rule->{-inpatterns};
    return unless ($patternptr);
    my $rules=$rule->{-rules};
    my @nvr=$rules->get_nvr($rule->specfile);
    return unless @nvr;
    my @gitpaths;
    foreach my $pattern (@$patternptr) {
	my $commit=$pattern->[0];
	my $patt0=$pattern->[0];
	if ($pattern->[0] ne '.') {
	    $patt0=&Gear::Rules::subst_gear_nvr($pattern->[0],@nvr);
	    $commit=$rules->get_gear_tag_commit($patt0);
	    unless ($commit) {
		warn "line ".$rule->{-line}.": tree_path $pattern->[0]:$pattern->[1]: can't resolve tag $patt0. In ".$rule->{-rulestring}."\n\n";
		die 'Error: .gear/tags/list is in inconsistent state.
The branch/tag git expression ',$patt0,' is not present in .gear/tags/list.
Run gear-update-tags -a and commit .gear/tags/list
to synchronize the version in spec file and branches/tags in .gear/tags/list',"\n";
	    }
	}
	my $patt1=&Gear::Rules::subst_gear_nvr($pattern->[1],@nvr);
	push @gitpaths, {
	    'commit'	=> $commit,
	    'tag'	=> $patt0,
	    'path'	=> $patt1,
	    'tag_template'	=> $pattern->[0],
	    'path_template'	=> $pattern->[1],
	};
    }
    return @gitpaths;
}


sub describe {
    my $rule=shift;
    return "line ".$rule->{-line}.": ".$rule->{-rulestring};
}

package Gear::Git::Helper;
# public static calls

sub get_git_toplevel {
    #return `git rev-parse --show-cdup`;
    open (my $fn, '-|', 'git rev-parse --show-cdup') || die "can't exec git rev-parse --show-cdup";
    my $path=<$fn>;
    close ($fn) or die "git rev-parse --show-cdup failed.\n";
    chomp $path;
    return $path;
}

sub get_current_git_branch {
    open (my $fn, 'git branch |') or die "can't run git branch: $!";
    my $current_git_branch;
    while (<$fn>) {
	chomp;
	$current_git_branch=$1 if /^\*\s+(\S.*)$/;
    }
    close($fn);
    #return &get_git_commit('HEAD') if $current_git_branch eq '(no branch)';
    return undef if $current_git_branch eq '(no branch)';
    return $current_git_branch;
}

sub get_git_commit {
    my ($name)=@_;
    my $commit=`git show --format=short | head -1`;
    die "git show failed\n" if $?;
    chomp $commit;
    die "unrecognized reply of git show: $commit\n" unless $commit=~s/^commit\s+//;
    return $commit;
}

sub ls_git_branches {
    my ($contains)=@_;
    open (my $fn, 'git branch '.($contains ? "--contains $contains" : '').' |') or die "can't run git branch: $!";
    my @git_branch;
    while (<$fn>) {
	chomp;
	s/^\*?\s+//;;
	push @git_branch, $_;
    }
    close($fn);
    return @git_branch;
}

sub get_git_branch_commit_hash {
    my ($contains)=@_;
    open (my $fn, 'git branch -v --no-abbrev '.($contains ? "--contains $contains" : '').' |') or die "can't run git branch: $!";
    my %git_branch;
    while (<$fn>) {
	chomp;
	die "Oops! Can't parse git branch -v --no-abbrev output: $_" unless s/^\*?\s+(\S+)\s+(\S+)//;;
	$git_branch{$1}=$2;
    }
    close($fn);
    return \%git_branch;
}

sub ls_git_tags {
    open (my $fn, 'git tag |') or die "can't run git tag: $!";
    my @git_tag;
    while (<$fn>) {
	chomp;
	push @git_tag, $_;
    }
    close($fn);
    return @git_tag;
}

1;

=head1	NAME

Gear::Rules - ALTLinux Gear::Rules parsing library.

=head1	DESCRIPTION

Gear::Rules - ALTLinux Gear::Rules parsing library.

=head1	AUTHOR

Written by Igor Vlasenko <viy@altlinux.org>.

=head1	COPYING

Copyright (c) 2010-2019 Igor Vlasenko, ALT Linux Team.

This is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

=cut

